/* High Level TDC V1190 Driver - MG xx/xx/2005 @ LIXAM-CNRS*/

/* --------------------------------------------------------------------------------------------------------------------- */
/* INCLUDES */
#include <stdio.h>
#include <stdlib.h>
#include "tdc1190_driver_high.h"

/* --------------------------------------------------------------------------------------------------------------------- */
/* DEFINES */
#define MAX_NUM_CHANNEL         0x007F

#define ACQUSITION_MODE         0x0001
#define TRIG_WINDOW             0x0FFF
#define TRIG_OFFSET_S           0x0800
#define TRIG_OFFSET_V           0x07FF
#define TRIG_CONFIG             0x0001
#define EVENT_MAX_HITS          0x000F
#define EVENT_HEADER_TRAILER    0x0003
#define EDGE_CONFIGURATION      0x0003
#define EDGE_RESOLUTION_1       0x0003
#define EDGE_RESOLUTION_2       0x0F07
#define CHANNEL_DEAD_TIME       0x0003
#define CHANNEL_OFFSET          0x00FF
#define HEADER_STATUS           0x0003
#define TDC_ERROR_STATUS        0x07FF
#define TDC_FIFO                0x0007
#define TDC_OFFSET_COARSE       0x07FF
#define TDC_OFFSET_FINE         0x001F

#define CANNOT_FIND_MAXHIT      TDC_ERROR|0x10
#define CANNOT_FIND_EDGE        TDC_ERROR|0x11
#define CANNOT_FIND_DEADTIME    TDC_ERROR|0x12
#define CANNOT_FIND_FIFOSIZE    TDC_ERROR|0x13

/* --------------------------------------------------------------------------------------------------------------------- */
/* CONSTANT */
int MaxHit[11]={10,0,1,2,4,8,16,32,64,128,65536}; /* Nb hits */
int Edge_Res[4]={3,800,200,100}; /* ps unit */
int Edge_Pair[9]={8,100,200,400,800,1600,3120,6250,12500}; /* ps unit */
int Edge_Width[15]={14,100,200,400,800,1600,3120,6250,12500,25000,50000,100000,200000,400000,800000}; /* ps unit */
int Channel_DeadTime[5]={4,5,10,30,100}; /* ns unit */
int FifoSize[9]={8,2,4,8,16,32,64,128,256}; /* 32-bit word unit */


/* --------------------------------------------------------------------------------------------------------------------- */
/* PROTOYPE */
unsigned int Find_InTable(int *table, unsigned int value);
int Get_SignedNumber(int value,int hsb,int lsb);
int Set_SignedNumber(int value,int hsb,int lsb);

/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* READ FROM TDC MICRO CONTROLLER */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
int GetTdc_AcqMode(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_ACQ_MOD;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->Acquisition_Mode=(char)(op.operand[0]&ACQUSITION_MODE);
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_TrigConfig(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_TRG_CONF;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;

  a_tdc->Trigger_Config[0]=(op.operand[0]&TRIG_WINDOW);                    /* window width - 12 bits*/
  a_tdc->Trigger_Config[1]=(op.operand[1]&TRIG_WINDOW);                    /* window offset- bit[11]-> sign; bit[0:10]->value*/
  a_tdc->Trigger_Config[2]=(op.operand[2]&TRIG_WINDOW);                    /* margin extra search - 12 bits*/
  a_tdc->Trigger_Config[3]=(op.operand[3]&TRIG_WINDOW);                    /* margin reject - 12 bits*/
  a_tdc->Trigger_Config[4]=(op.operand[4]&TRIG_CONFIG);                    /* trigger substraction enable ?*/
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EdgeConfig(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_DETECTION;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->Edge_Config=(char)(op.operand[0]&EDGE_CONFIGURATION);
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EdgeRes(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_RES;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  if (a_tdc->Edge_Config)
    a_tdc->Edge_Resolution=(op.operand[0]&EDGE_RESOLUTION_1);
  else
    a_tdc->Edge_Resolution=(op.operand[0]&EDGE_RESOLUTION_2);
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventMaxHit(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_EVENT_SIZE;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->Event_MaxHit=(char)(op.operand[0]&EVENT_MAX_HITS);
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventHeaderTrailer(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_HEAD_TRAILER;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->Event_HeaderTrailer=(char)(op.operand[0]&EVENT_HEADER_TRAILER);
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_ChannelDeadTime(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_DEAD_TIME;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->Channel_DeadTime=(char)(op.operand[0]&CHANNEL_DEAD_TIME);
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_ChannelActive(tdc1190 *a_tdc)
{
  int res,i;
  opcode op;
  op.command=READ_EN_PATTERN;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  for (i=0; i<op.nR;i++)
    a_tdc->Channel_Active[i]=op.operand[i];
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_ChanNumOffset(tdc1190 *a_tdc, int NumChannel)
{
  int res;
  opcode op;
  op.command=READ_ADJUST_CH;
  op.object=(char)(NumChannel)&(a_tdc->NumberOfChannel);
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->Channel_Offset[NumChannel]=(char)(op.operand[0]&CHANNEL_OFFSET);
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_ErrorTypes(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_ERROR_TYPES;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->TDC_ErrorTypes=(op.operand[0]&TDC_ERROR_STATUS);
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_FifoSize(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_FIFO_SIZE;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->TDC_FifoSize=(char)(op.operand[0]&TDC_FIFO);
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_GlobalOffset(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=READ_GLOB_OFFS;
  res=read_micro(&a_tdc->tdc_access,&op); if (res) return res;
  a_tdc->TDC_TimeOffset[0]=(op.operand[0]& TDC_OFFSET_COARSE);
  a_tdc->TDC_TimeOffset[1]=(op.operand[1]& TDC_OFFSET_FINE);
  return NO_TDC_ERROR;
}

/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* WRITE TO TDC MICRO CONTROLLER */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
int SetTdc_AcqMode(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value -> Trigger Matching */
    op.command=TRG_MATCH;
  else
    op.command=CONT_STOR;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_AcqMode(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_KeepToken(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value -> Keep Token */
    op.command=SET_KEEP_TOKEN;
  else 
    op.command=CLEAR_KEEP_TOKEN;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_LoadConfig(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* default true value -> user */
    op.command=LOAD_USER_CONFIG;
  else
    op.command=LOAD_DEF_CONFIG;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  res=GetTdc_AcqMode(a_tdc); if (res) return res;
  res=GetTdc_TrigConfig(a_tdc); if (res) return res;
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_SaveUserConfig(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=SAVE_USER_CONFIG;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_AutoLoadConfig(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* default true value -> user */
    op.command=AUTOLOAD_USER_CONF;
  else
    op.command=AUTOLOAD_DEF_CONF;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_Window(tdc1190 *a_tdc, unsigned int width, int offset)
{
  int res;
  opcode op;
  op.command=SET_WIN_WIDTH;
  op.operand[0]=((width/25)& TRIG_WINDOW); /* 25 ns base time*/
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  op.command=SET_WIN_OFFS;
  op.operand[0]=Set_SignedNumber(offset/25, TRIG_OFFSET_S, TRIG_OFFSET_V); /* 25 ns base time; sign may not work*/
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_TrigConfig(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_Margin(tdc1190 *a_tdc, unsigned int extrasearch, unsigned int reject)
{
  int res;
  opcode op;
  op.command=SET_SW_MARGIN;
  op.operand[0]=((extrasearch/25)& TRIG_WINDOW); /* 25 ns base time*/
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  op.command=SET_REJ_MARGIN;
  op.operand[0]=((reject/25)& TRIG_WINDOW); /* 25 ns base time */
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;

  return GetTdc_TrigConfig(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_TriggerSubstraction(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /*Default true value -> enable trigger time substraction */
    op.command=EN_SUB_TRG; 
  else
    op.command=DIS_SUB_TRG;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_TrigConfig(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_EdgeConfig(tdc1190 *a_tdc, unsigned int edge)
{
  int res;
  opcode op;
  op.command=SET_DETECTION;
  op.operand[0]=(edge&EDGE_CONFIGURATION);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_EdgeConfig(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_EdgeRes(tdc1190 *a_tdc, unsigned int edge, unsigned int width)
{
  int res,dummy1, dummy2;
  opcode op;
  if (a_tdc->Edge_Config)
    {
      dummy1=Find_InTable(Edge_Res,edge); if(dummy1==-1) return CANNOT_FIND_EDGE;
      op.command=SET_TR_LEAD_LSB;
      op.operand[0]=(dummy1&EDGE_RESOLUTION_1);
    }
  else
    {
      dummy1=Find_InTable(Edge_Width,width);if(dummy1==-1) return CANNOT_FIND_EDGE;
      dummy2=Find_InTable(Edge_Pair,edge);;if(dummy2==-1) return CANNOT_FIND_EDGE;
      op.command=SET_PAIR_RES;
      op.operand[0]=(((dummy1<<8)+dummy2)&EDGE_RESOLUTION_2);
    }
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_EdgeRes(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_EventMaxHit(tdc1190 *a_tdc, int NbHits)
{
  int res,dummy;
  opcode op;
  op.command=SET_EVENT_SIZE;
  dummy=Find_InTable(MaxHit,NbHits);if(dummy==-1) return CANNOT_FIND_MAXHIT;
  op.operand[0]=dummy&EVENT_MAX_HITS;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_EventMaxHit(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_EventHeaderTrailer(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value -> enable header and trailer */
    op.command=EN_HEAD_TRAILER;
  else
    op.command=DIS_HEAD_TRAILER;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_EventHeaderTrailer(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ChanDeadTime(tdc1190 *a_tdc, int deadtime)
{
  int res;
  opcode op;
  op.command=SET_DEAD_TIME;
  deadtime=Find_InTable(Channel_DeadTime, deadtime); if (deadtime==-1) return CANNOT_FIND_DEADTIME;
  op.operand[0]=(deadtime&CHANNEL_DEAD_TIME);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_ChannelDeadTime(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ChannelActive(tdc1190 *a_tdc, unsigned int *table)
{
  int res,i;
  opcode op;
  op.command=WRITE_EN_PATTERN;
  for (i=0; i<8;i++)
    op.operand[i]=*(table+i);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_ChannelActive(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_AllChannelActive(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value ->  enable all channel*/
    op.command=EN_ALL_CH;
  else
    op.command=DIS_ALL_CH;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ChanNumActive(tdc1190 *a_tdc, int NumChannel, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value -> enable channel NumChannel */
    op.command=EN_CHANNEL;
  else
    op.command=DIS_CHANNEL;
  op.object=(char)(NumChannel)&(a_tdc->NumberOfChannel);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ChanNumOffset(tdc1190 *a_tdc, int NumChannel, unsigned int offset)
{
  int res;
  opcode op;
  op.command=SET_ADJUST_CH;
  op.object=(char)(NumChannel)&(a_tdc->NumberOfChannel);
  op.operand[0]=(offset&CHANNEL_OFFSET);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_ChanNumOffset(a_tdc,NumChannel);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ErrorMark(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value -> enable error mark */
    op.command=EN_ERROR_MARK;
  else
    op.command=DIS_ERROR_MARK;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ErrorBypass(tdc1190 *a_tdc, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value -> enable error mark */
    op.command=EN_ERROR_BYPASS;
  else
    op.command=DIS_ERROR_BYPASS;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ErrorTypes(tdc1190 *a_tdc)
{
  int res;
  opcode op;
  op.command=SET_ERROR_TYPES;
  op.operand[0]= (a_tdc->TDC_ErrorTypes&TDC_ERROR_STATUS);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_ErrorTypes(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_FifoSize(tdc1190 *a_tdc, int size)
{
  int res;
  opcode op;
  op.command=SET_FIFO_SIZE;
  size=Find_InTable(FifoSize,size);if (size==-1) return CANNOT_FIND_FIFOSIZE;
  op.operand[0]=size&TDC_FIFO;
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_FifoSize(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_GlobalOffset(tdc1190 *a_tdc, unsigned int coarse, unsigned int fine)
{
  int res;
  opcode op;
  op.command=SET_GLOB_OFFS;
  op.operand[0]=(coarse&TDC_OFFSET_COARSE);
  op.operand[1]=(fine&TDC_OFFSET_FINE);
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return GetTdc_GlobalOffset(a_tdc);
}

/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_TestMode(tdc1190 *a_tdc, unsigned int lsb_testword, unsigned int hsb_testword, int status)
{
  int res;
  opcode op;
  if (status) /* Default true value ->  enable all channel*/
    {
      op.command=ENABLE_TEST_MODE;
      op.operand[0]=lsb_testword;
      op.operand[1]=hsb_testword;
    }
  else
    op.command=DISABLE_TEST_MODE;
  
  res=write_micro(&a_tdc->tdc_access,&op); if (res) return res;
  return NO_TDC_ERROR;
}
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* READ FROM TDC REGISTERS */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
int GetTdc_Status(tdc1190 *a_tdc,unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,STATUS_REGISTER,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_Event(tdc1190 *a_tdc, unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,OUTPUT_REGISTER,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventCounter(tdc1190 *a_tdc, unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,EVENT_COUNTER,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventStored(tdc1190 *a_tdc, unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,EVENT_STORED,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_Firmware(tdc1190 *a_tdc,unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,FIRMWARE_VERSION,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventFIFO(tdc1190 *a_tdc,unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,EVENT_FIFO,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventFIFO_Stored(tdc1190 *a_tdc,unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,EVENT_FIFO_STORED,value);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_EventFIFO_Status(tdc1190 *a_tdc,unsigned int *value)
{
  int res;
  res=read_register(&a_tdc->tdc_access,EVENT_FIFO_STATUS,value);if (res) return res;
  return NO_TDC_ERROR;
}



/* --------------------------------------------------------------------------------------------------------------------- */

int GetTdc_Control(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,CONTROL_REGISTER,&a_tdc->CtrlReg);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */

int GetTdc_InterruptLevel(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,INTERRUPT_LEVEL,&a_tdc->IrqLevel);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_InterruptVector(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,INTERRUPT_VECTOR,&a_tdc->IrqVector);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_GEO_Address(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,GEO_ADDRESS_REGISTER,&a_tdc->GeoAddress);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_MCST_Address(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,MCST_BASE_ADDRESS,&a_tdc->MCSTAddress);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_MCST_Control(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,MCST_CONTROL,&a_tdc->MCSTControl);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */

int GetTdc_AllmostFull(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,ALMOST_FULL_EVENT,&a_tdc->AlmostFull);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_BLT_EventNumber(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,BTL_EVENT_NUMBER,&a_tdc->BLTEventNumber);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_TestReg(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,TEST_REG,&a_tdc->TestReg);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_OutProg_Control(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,OUTPUT_PROG_CONTROL,&a_tdc->OutProg);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_Dummy32(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,DUMMY32,&a_tdc->Dummy32);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int GetTdc_Dummy16(tdc1190 *a_tdc)
{
  int res;
  res=read_register(&a_tdc->tdc_access,DUMMY16,&a_tdc->Dummy16);if (res) return res;
  return NO_TDC_ERROR;
}
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* WRITE TO TDC REGISTERS */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
int SetTdc_ModuleReset(tdc1190 *a_tdc)
{
  int res;
  res=write_register(&a_tdc->tdc_access,MODULE_RESET,0);if (res) return res;
  res=InitTdc(a_tdc,a_tdc->tdc_access.address_device,a_tdc->NumberOfChannel);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ClearRegister(tdc1190 *a_tdc)
{
  int res;
  res=write_register(&a_tdc->tdc_access,SOFTWARE_CLEAR,0);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_ClearEvent(tdc1190 *a_tdc)
{
  int res;
  res=write_register(&a_tdc->tdc_access,SOFTWARE_EVENT_RESET,0);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_SoftTrigger(tdc1190 *a_tdc)
{
  int res;
  res=write_register(&a_tdc->tdc_access,SOFTWARE_TRIGGER,0);if (res) return res;
  return NO_TDC_ERROR;
}
/* --------------------------------------------------------------------------------------------------------------------- */

int SetTdc_Control(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,CONTROL_REGISTER,value);if (res) return res;
  return GetTdc_Control(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */

int SetTdc_InterruptLevel(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,INTERRUPT_LEVEL,value);if (res) return res;
  return GetTdc_InterruptLevel(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_InterruptVector(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,INTERRUPT_VECTOR,value);if (res) return res;
  return GetTdc_InterruptVector(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_GEO_Address(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,GEO_ADDRESS_REGISTER,value);if (res) return res;
  return GetTdc_GEO_Address(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_MCST_Address(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,MCST_BASE_ADDRESS,value);if (res) return res;
  return GetTdc_MCST_Address(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_MCST_Control(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,MCST_CONTROL,value);if (res) return res;
  return GetTdc_MCST_Control(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */

int SetTdc_AlmostFull(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,ALMOST_FULL_EVENT,value);if (res) return res;
  return GetTdc_AllmostFull(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_BLT_EventNumber(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,BTL_EVENT_NUMBER,value);if (res) return res;
  return GetTdc_BLT_EventNumber(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_TestReg(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,TEST_REG,value);if (res) return res;
  return GetTdc_TestReg(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_OutProg_Control(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,OUTPUT_PROG_CONTROL,value);if (res) return res;
  return GetTdc_OutProg_Control(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_Dummy32(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,DUMMY32,value);if (res) return res;
  return GetTdc_Dummy32(a_tdc);
}
/* --------------------------------------------------------------------------------------------------------------------- */
int SetTdc_Dummy16(tdc1190 *a_tdc, unsigned int value)
{
  int res;
  res=write_register(&a_tdc->tdc_access,DUMMY16,value);if (res) return res;
  return GetTdc_Dummy16(a_tdc);
}
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* STRUCTURE ACCESS */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
int InitTdc(tdc1190 *a_tdc, unsigned int address, unsigned char NbChannel)
{
  int res;
  
  open_vme(&a_tdc->tdc_access,address);
  a_tdc->tdc_access.address_device=address;
  a_tdc->tdc_access.fd=0;
  a_tdc->NumberOfChannel=NbChannel;
  /* Registers*/
  res=GetTdc_Control(a_tdc); if (res) return res;
  res=GetTdc_InterruptLevel(a_tdc); if (res) return res;
  res=GetTdc_InterruptVector(a_tdc); if (res) return res;
  res=GetTdc_GEO_Address(a_tdc); if (res) return res;
  res=GetTdc_MCST_Address(a_tdc); if (res) return res;
  res=GetTdc_MCST_Control(a_tdc); if (res) return res;
  res=GetTdc_AllmostFull(a_tdc); if (res) return res;
  res=GetTdc_BLT_EventNumber(a_tdc); if (res) return res;
  res=GetTdc_TestReg(a_tdc); if (res) return res;
  res=GetTdc_OutProg_Control(a_tdc); if (res) return res;
  res=GetTdc_Dummy32(a_tdc); if (res) return res;
  res=GetTdc_Dummy16(a_tdc); if (res) return res;
  /* Micro Controler */
  res=GetTdc_AcqMode(a_tdc); if (res) return res;
  res=GetTdc_TrigConfig(a_tdc);if (res) return res;
  res=GetTdc_EdgeConfig(a_tdc);if (res) return res;
  res=GetTdc_EdgeRes(a_tdc);if (res) return res;
  res=GetTdc_EventMaxHit(a_tdc);if (res) return res;
  res=GetTdc_EventHeaderTrailer(a_tdc);if (res) return res;
  res=GetTdc_ChannelDeadTime(a_tdc);if (res) return res;
  res=GetTdc_ChannelActive(a_tdc);if (res) return res;
  res=GetTdc_ChanNumOffset(a_tdc, 1);if (res) return res;
  res=GetTdc_ErrorTypes(a_tdc);if (res) return res;
  res=GetTdc_FifoSize(a_tdc);if (res) return res;
  res=GetTdc_GlobalOffset(a_tdc);if (res) return res;


 return NO_TDC_ERROR; 

}
/* --------------------------------------------------------------------------------------------------------------------- */
void PrintTdc_RegisterStatus(tdc1190 *a_tdc)
{
  unsigned int value;

  printf("========================== REGISTER STATUS =========================== \n");
  printf("Control Register :bx%9.0f /  ",Set2Bit(a_tdc->CtrlReg));
  GetTdc_Status(a_tdc, &value);
  printf("Status Register :bx%15.0f /  ",Set2Bit(value));
  printf("OUT_PROG Register:bx%3.0f\n",Set2Bit(a_tdc->OutProg));  
  printf("Allmost full register Level :0x%04x  /  ",a_tdc->AlmostFull);
  printf("Irq Level:0x%04x  /  ",a_tdc->IrqLevel);
  printf("Irq Vector:0x%04x\n",a_tdc->IrqVector); 
  printf("GEO Address:0x%04x  /  ",a_tdc->GeoAddress);
  printf("MCST/CBLT Control:0x%04x  /  ",a_tdc->MCSTControl);  
  printf("MCST Base Address:0x%04x\n",a_tdc->MCSTAddress);
  printf("BLT Event Number:0x%04x\n",a_tdc->BLTEventNumber);
  printf("Test Register:0x%08x\n",a_tdc->TestReg);
  printf("Dummy 32:0x%08x  /  ",a_tdc->Dummy32);
  printf("Dummy 16:0x%04x\n\n",a_tdc->Dummy16);

}

/* --------------------------------------------------------------------------------------------------------------------- */
void PrintTdc_MicroStatus(tdc1190 *a_tdc)
{
  int i;
  printf("=================== MICRO CONTROLLER OPERATION  STATUS ======================== \n");
  if (a_tdc->Acquisition_Mode) printf("Acquisition Mode :Trigger Matching Mode\n");
  else printf("Acquisition Mode: Continuous Storage Mode\n");
  if (a_tdc->Trigger_Config[4]) printf("Trigger substraction: enable\n");
  else  printf("Trigger substraction: disable\n");
  printf("Trigger Window - Width: %i ns / ",a_tdc->Trigger_Config[0]*25);
  printf("Offset: %i ns  ",Get_SignedNumber(a_tdc->Trigger_Config[1],TRIG_OFFSET_S,TRIG_OFFSET_V)*25);
  printf("(0x%04x)\n",a_tdc->Trigger_Config[1]);
  printf("Trigger Margin - Extra search: %i ns / ",a_tdc->Trigger_Config[2]*25);
  printf("Reject: %i ns\n",a_tdc->Trigger_Config[3]*25);
  printf("-------------------------------------\n");
  printf("Event - Nb max of hits: %i / ",MaxHit[a_tdc->Event_MaxHit+1]);
  if (a_tdc->Event_HeaderTrailer) printf("Header & Trailer: enable\n");
  else  printf("Header & Trailer: disable\n");
  printf("-------------------------------------\n");
  switch (a_tdc->Edge_Config) {
  case 0: printf("Measurement: Pair  - "); break;
  case 1: printf("Measurement: Trailing  - "); break;
  case 2: printf("Measurement: Leading  - "); break;
  case 3: printf("Measurement: Trailing and Leading  - "); break;
  }
  if (a_tdc->Edge_Config)
    printf("Resolution: %i ps\n",Edge_Res[a_tdc->Edge_Resolution+1]);
  else {
    printf("Resolution: %i ps / ",Edge_Pair[(a_tdc->Edge_Resolution&0x0007)+1]);
    printf("Width: %i ps\n",Edge_Width[((a_tdc->Edge_Resolution&0x0F00)>>8)+1]);
    }
  printf("-------------------------------------\n");
  printf("Channel Dead Time: %i (ns)\n",Channel_DeadTime[a_tdc->Channel_DeadTime+1]);
  for (i=0;i<8;i+=2) 
    printf("Channel Pattern %i : 0x%04x    Channel Pattern %i : 0x%04x\n",
	   i, a_tdc->Channel_Active[i],i+1, a_tdc->Channel_Active[i+1]);
  printf("-------------------------------------\n");
  printf("TDC Error Types : bx%11.0f \n",Set2Bit(a_tdc->TDC_ErrorTypes));
  printf("TDC Fifo Size: %i words\n",FifoSize[a_tdc->TDC_FifoSize+1]);
  printf("TDC Global Time Offset (coarse): %i  ",a_tdc->TDC_TimeOffset[0]);
  printf("(fine): %i\n\n",a_tdc->TDC_TimeOffset[1]);

}
/* --------------------------------------------------------------------------------------------------------------------- */
void PrintTdc_Status(tdc1190 *a_tdc)
{
  PrintTdc_RegisterStatus(a_tdc);
  PrintTdc_MicroStatus(a_tdc);
}

/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* MISCELLANEOUS*/
/* ============================================================================================================================ */
/* ============================================================================================================================ */
/* ============================================================================================================================ */
unsigned int Find_InTable(int *table, unsigned int value)
{
  int i=1,not_ok=1, max_value;

  max_value=*(table)+1;
  do {
    /*printf("%i <-> %i\n",*(table+i),value);*/
    if (*(table+i)==value) not_ok=0;
    i+=1;
  }
  while((not_ok)&&(i<max_value));
  if (i>max_value) return -1;
  else return i-2;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int Get_SignedNumber(int value,int hsb,int lsb)
{ 
  int sign=1;
  if (value&hsb) sign=-1;
  return ((value*sign)&lsb)*sign;
}
/* --------------------------------------------------------------------------------------------------------------------- */
int Set_SignedNumber(int value,int hsb,int lsb)
{ 
  int sign=1;
  if (value<0) {
    value=-value;
    sign=-1;
  }
  return (value&lsb)*sign;
}
/* --------------------------------------------------------------------------------------------------------------------- */
double Set2Bit(unsigned int value)
{ 
  char s[33];
  int i;
  s[32]=0;
  for(i=0;i<32;i++){
      if ((value>>i)&1) s[31-i]='1';
      else s[31-i]='0';
  }
  /*printf("bin2string : %s, %llu\n",s, (unsigned long long)atof(s));*/
  return atof(s);
}
